/*
 * Index tracking example
 *
 * See the attached report index_tracking.pdf for the description of the model
 * and for the compilation instructions.
 * NAG Ltd. 2016
 *
 * Example for Linux.
 */

#include <nag.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <nag_stdlib.h>
#include <nage04.h>
#include <nagf16.h>


/* dynamically allocate a 2-dimensional array mxn, type double, return 0 on failure */
double** allocate_2d(int m, int n)
{
	double **arr = (double **)malloc(m * sizeof(double *) + (m * n * sizeof(double)));
	if (arr) {
		double *mem = (double *)(arr + m);

		for (int i = 0; i < m; i++)
			arr[i] = mem + (i * n);
	}
	return arr;
}

/* dynamically allocate array of char, return 0 on failure */
char** allocate_2d_char(int m, int n)
{
	char **arr = (char **)malloc(m * sizeof(char *) + (m * n * sizeof(char)));
	if (arr) {
		char *mem = (char *)(arr + m);

		for (int i = 0; i < m; i++)
			arr[i] = mem + (i * n);
	}
	return arr;
}

/* allocate 1-dimensional array, type double, return 0 on failure */
double* allocate_1d(int n)
{
	double *a = (double *)malloc(n * sizeof(double));
	return a;
}

/* allocate 1-dimensional array, type int, return 0 on failure */
Integer* allocate_1d_int(int n)
{
	Integer *a = (Integer *)malloc(n * sizeof(Integer));
	return a;
}

/* computes arithmetic mean of the array of n elements */
double mean(double arr[], int n)
{
	double sum = 0.0;
	for (int i = 0; i < n; i++)
		sum += arr[i];

	return sum / n;
}

/* computes arithmetic mean by columns of the matrix mxn elements */
void mean_2d(int m, int n, double** arr, double* means)
{
	for (int i = 0; i < n; i++)
	{
		means[i] = 0.0;
		for (int j = 0; j < m; j++)
			means[i] += arr[j][i];
		means[i] /= m;
	}
}

/* estimates standard deviation based on a sample of n elements (divides by n - 1 elements) */
double stdev_s(double arr[], int n)
{
	double mu = mean(arr, n);
	double sum = 0.0;
	for (int i = 0; i < n; i++)
		sum += pow(arr[i] - mu, 2);

	return sqrt(sum / (double)(n - 1));
}

/* estimates standard deviation based on a sample for n-variable and m observations (divides by m - 1 elements)
 * returns 0 if failed otherwise 1
 */
int stdev_s_2d(int m, int n, double** arr, double* stdevs)
{
	double *mu;
	mu = allocate_1d(n);
	if (!mu)
		return 0;
	mean_2d(m, n, arr, mu);

	for (int i = 0; i < n; i++)
	{
		stdevs[i] = 0.0;
		for (int j = 0; j < m; j++)
			stdevs[i] += pow(arr[j][i] - mu[i], 2);

		stdevs[i] = sqrt(stdevs[i] / (double)(m - 1));
	}
	free(mu);
	return 1;
}

/* estimates sample covariance given two arrays with the same dimension n */
double covariance(int n, double* arr1, double* arr2)
{
	double mu1, mu2, sum;
	mu1 = mean(arr1, n);
	mu2 = mean(arr2, n);

	sum = 0.0;
	for (int i = 0; i < n; i++)
		sum += (arr1[i] - mu1) * (arr2[i] - mu2);

	return sum / (double)(n - 1);
}

/* estimates sample covariance matrix given an array of mxn elements
 * returns 0 if failed otherwise 1
 */
int covariance_matrix(int m, int n, double** arr, double** cov)
{
	double *mu, sum;
	mu = allocate_1d(n);
	if (!mu)
		return 0;
	mean_2d(m, n, arr, mu);

	for (int i = 0; i < n; i++)
		for (int j = i; j < n; j++)
		{
			sum = 0.0;
			for (int k = 0; k < m; k++)
				sum += (arr[k][i] - mu[i]) * (arr[k][j] - mu[j]);
			cov[i][j] = sum / (double)(m - 1);
			cov[j][i] = cov[i][j];
		}
	free(mu);
	return 1;
}

/* computes the beta of n-assets with m observations relative to the market index and the coefficients of the linear term */
void betas(int m, int n, double* index, double index_mu, double index_std, double** asset, double* asset_mu, double* b, double* cvec)
{
	double sum;

	for (int i = 0; i < n; i++)
	{
		sum = 0.0;
		for (int k = 0; k < m; k++)
			sum += (asset[k][i] - asset_mu[i]) * (index[k] - index_mu);
		
		cvec[i] = -sum / (double)(m - 1);
		b[i] = -cvec[i] / (index_std * index_std);
	}
}

int main(void)
{
	Integer exit_status = 0;
	Integer i, j, *kx=0, n, nclin, rec, std_err, cov_err;
	double objf;
	double **a=0, *asset_mean=0, **assets_r=0, *asset_stdv=0, *b=0, *bl=0, *bu=0, *cvec=0, **cov=0, **cov2=0, **data=0, *index_r=0, *x=0, *y=0;
	double index_mean, index_stdv, portf_var, portf_beta, measure;
	char **tickers=0;

	/* Nag Types */
	Nag_E04_Opt options;
	Nag_Comm comm;
	NagError fail;
	Nag_OrderType order;
	Nag_UploType  uplo;

	INIT_FAIL(fail);

	printf("Index tracking example\n");
	fflush(stdout);

	/* read problem dimensions: number of assets (plus market index) and number of data points  */
	scanf("%" NAG_IFMT "%" NAG_IFMT "%*[^\n]", &n, &rec);

	/* allocate assets data */
	tickers = allocate_2d_char(n + 1, 5);
	data = allocate_2d(rec, n + 1);
	assets_r = allocate_2d(rec - 1, n);
	index_r = allocate_1d(rec - 1);
	asset_mean = allocate_1d(n);
	asset_stdv = allocate_1d(n);
	cov = allocate_2d(n, n);
	cov2 = allocate_2d(n, n);
	b = allocate_1d(n);
	
	/* allocate solver data*/
	nclin = 2;
	a = allocate_2d(nclin, n);
	bl = allocate_1d(nclin + n);
	bu = allocate_1d(nclin + n);
	cvec = allocate_1d(n);
	x = allocate_1d(n);
	y = allocate_1d(n);
	kx = allocate_1d_int(n);

	if (!data || !assets_r || !index_r || !asset_mean || !asset_stdv || !cov || ! cov2 || !b
		|| !a || !bl || !bu || !cvec || !x || !kx || !tickers || !y)
	{
		printf("Allocation failure\n");
		exit_status = -1;
		goto END;
	}

	/* read tickers */
	for (i = 0; i < n + 1; i++)
		scanf("%4s", &tickers[i][0]);
		
	/* read data */
	for (i = 0; i < rec; ++i)
		for (j = 0; j < n + 1; ++j)
			scanf("%lf", &data[i][j]);

	/* compute one period returns */
	for (i = 0; i < n + 1; i++)
		for (j = 0; j < rec - 1; j++)
		{
			if (i == 0)
				index_r[j] = (data[j][i] - data[j + 1][i]) / data[j + 1][i];
			else
				assets_r[j][i - 1] = (data[j][i] - data[j + 1][i]) / data[j + 1][i];
		}

	/* compute mean of returns */
	index_mean = mean(index_r, rec - 1);
	mean_2d(rec - 1, n, assets_r, asset_mean);
	
	/* compute standard deviation of returns */
	index_stdv = stdev_s(index_r, rec - 1);
	std_err = stdev_s_2d(rec - 1, n, assets_r, asset_stdv);
	if (!std_err)
	{
		printf("Std. deviation failed\n");
		exit_status = -1;
		goto END;
	}

	/* compute covariance matrix */
	cov_err = covariance_matrix(rec - 1, n, assets_r, cov);
	if (!cov_err)
	{
		printf("Covariance matrix failed\n");
		exit_status = -1;
		goto END;
	}

	/* compute portfolio betas */
	betas(rec - 1, n, index_r, index_mean, index_stdv, assets_r, asset_mean, b, cvec);

	/* matrix of constraint A */
	for (i = 0; i < nclin; i++)
		for (j = 0; j < n; j++)
		{
			if (i == 0)
				a[i][j] = asset_mean[j];
			else
				a[i][j] = 1.0;
		}

	/* bounds bl and bu*/
	for (i = 0; i < n; i++)
	{
		bl[i] = -1.0;
		bu[i] = 1.0;
	}
	bl[n] = index_mean;
	bu[n] = index_mean;
	bl[n + 1] = 1.0;
	bu[n + 1] = 1.0;

	/* initial point x */
	for (i = 0; i < n; i++)
		x[i] = 0.0;
		
	/* copy covariance matrix for computing portfolio measures*/
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			cov2[i][j] = cov[i][j];

	/* nag_opt_init (e04xxc) */
	/* initialization function for option setting */
	nag_opt_init(&options);
	options.prob = Nag_QP2;

	/* nag_opt_lin_lsq (e04ncc), see above. */
	nag_opt_lin_lsq(n, n, nclin, &a[0][0], n, bl, bu, cvec, (double *)0,
		&cov[0][0], n, kx, x, &objf, &options, &comm, &fail);
	if (fail.code != NE_NOERROR) 
	{
		printf("Error from nag_opt_lin_lsq (e04ncc).\n%s\n", fail.message);
		exit_status = 1;
	}

	/* print optimal portfolio selection */
	printf("\nOptimal portfolio selection:\n");
	printf("----------------------------\n");
	printf("%5s: market index\n", tickers[0]);
	for (i = 1; i < n+1; i++)
		printf("%5s: %6.2lf%%\n", tickers[i], 100.0*x[i-1]);

	/* compute portfolio measures */
	printf("\nPortfolio measures:\n");
	printf("----------------------------\n");

	/* compute portfolio variance */
	/* nag_dsymv (f16pcc).
	* Symmetric matrix-vector multiply.
	*/
	order = Nag_RowMajor;
	uplo = Nag_Upper;
	nag_dsymv(order, uplo, n, 1.0, &cov2[0][0], n, x, 1, 0.0, y, 1, &fail);
	for (portf_var = 0.0, i = 0; i < n; i++)
		portf_var += x[i] * y[i];

	printf("Portfolio variance : %lf\n", portf_var);
	
	/* compute portfolio beta */
	for (portf_beta = 0.0, i = 0; i < n; i++)
		portf_beta += b[i] * x[i];

	printf("Portfolio beta     : %lf\n", portf_beta);

	/* compute measure of goodness: Var(r_P - r_M) */
	measure = portf_var + (index_stdv * index_stdv) * (1 - 2 * portf_beta);
	printf("Measure of goodness: %lf\n", measure);

	/* free options memory */
	/* nag_opt_free (e04xzc).
	 * memory freeing function for use with option setting
	 */
	nag_opt_free(&options, "all", &fail);
	if (fail.code != NE_NOERROR) 
	{
		printf("Error from nag_opt_free (e04xzc).\n%s\n", fail.message);
		exit_status = 1;
		goto END;
	}

END:
	if (a) free(a);
	if (bl) free(bl);
	if (bu) free(bu);
	if (cvec) free(cvec);
	if (x) free(x);
	if (y) free(y);
	if (kx) free(kx);
	if (data) free(data);
	if (cov) free(cov);
	if (cov2) free(cov2);
	if (assets_r) free(assets_r);
	if (asset_mean) free(asset_mean);
	if (asset_stdv) free(asset_stdv);
	if (index_r) free(index_r);
	if (tickers) free(tickers);

	return exit_status;
}
